import 'dart:collection';
import 'dart:html' as html;
import 'package:path/path.dart' as Path;

import 'package:js/js.dart';
import 'package:js/js_util.dart';
import 'package:tencent_im_sdk_plugin_platform_interface/enum/V2TimAdvancedMsgListener.dart';
import 'package:tencent_im_sdk_plugin_platform_interface/models/v2_tim_callback.dart';
import 'package:tencent_im_sdk_plugin_platform_interface/models/v2_tim_message.dart';
import 'package:tencent_im_sdk_plugin_platform_interface/models/v2_tim_message_receipt.dart';
import 'package:tencent_im_sdk_plugin_platform_interface/models/v2_tim_message_search_result.dart';
import 'package:tencent_im_sdk_plugin_platform_interface/models/v2_tim_receive_message_opt_info.dart';
import 'package:tencent_im_sdk_plugin_platform_interface/models/v2_tim_value_callback.dart';
import 'package:tencent_im_sdk_plugin_web/src/enum/event_enum.dart';
import 'package:tencent_im_sdk_plugin_web/src/enum/group_receive_message_opt.dart';
import 'package:tencent_im_sdk_plugin_web/src/models/v2_tim_create_message.dart';
import 'package:tencent_im_sdk_plugin_web/src/models/v2_tim_get_message_list.dart';
import 'package:tencent_im_sdk_plugin_web/src/models/v2_tim_message.dart';
import 'package:tencent_im_sdk_plugin_web/src/utils/utils.dart';
import 'im_sdk_plugin_js.dart';
import 'package:mime_type/mime_type.dart';

class V2TIMMessageManager {
  static late TIM? timeweb;
  static V2TimAdvancedMsgListener? messageListener;

  V2TIMMessageManager() {
    timeweb = V2TIMManagerWeb.timWeb;
  }

  Future<dynamic> sendMessage<T, F>(
      {required String type, required Map<String, dynamic> params}) async {
    try {
      final groupID = params['groupID'] ?? '';
      final recveiver = params['receiver'] ?? '';
      final haveTwoValues = groupID != '' && recveiver != '';
      if (haveTwoValues) {
        return CommonUtils.returnErrorForValueCb<F>({
          'code': 6017,
          'desc': "receiver and groupID cannot set at the same time",
          'data': V2TimMessage(elemType: 1).toJson()
        });
      }
      final convType = groupID != '' ? 'GROUP' : 'C2C';
      final sendToUserID = convType == 'GROUP' ? groupID : recveiver;
      var messageElem;
      switch (type) {
        case "text":
          {
            final createElemParams = CreateMessage.createTextMessage(
                userID: sendToUserID,
                text: params["text"],
                convType: convType,
                priority: params['priority']);
            messageElem = timeweb!.createTextMessage(createElemParams);
            break;
          }
        case "custom":
          {
            final customMessage = CreateMessage.createCustomMessage(
                userID: sendToUserID,
                customData: params["data"],
                convType: convType,
                extension: params['extension'],
                description: params['desc'],
                priority: params['priority']);
            messageElem = timeweb!.createCustomMessage(customMessage);
            break;
          }
        case "face":
          {
            final faceMessage = CreateMessage.createFaceMessage(
                userID: sendToUserID,
                data: params["data"],
                index: params["index"],
                convType: convType,
                priority: params['priority']);
            messageElem = timeweb!.createFaceMessage(faceMessage);
            break;
          }
        case "image":
          {
            final progressCallback = allowInterop((progress) async {
              final messageInstance =
                  await V2TIMMessage.convertMessageFromWebToDart(messageElem);
              if (messageListener != null) {
                messageListener!.onSendMessageProgress(
                    V2TimMessage.fromJson(messageInstance), progress as int);
              }
            });
            final createElemParams = CreateMessage.createImageMessage(
                userID: sendToUserID,
                file: params['file'],
                convType: convType,
                progressCallback: progressCallback,
                priority: params['priority']);
            messageElem = timeweb!.createImageMessage(createElemParams);
            break;
          }
        case "textAt":
          {
            final createElemParams = CreateMessage.createTextAtMessage(
                groupID: sendToUserID,
                text: params['text'],
                priority: params['priority'],
                atList: params['atUserList']);
            messageElem = timeweb!.createTextAtMessage(createElemParams);
            break;
          }
        case "location":
          {
            final createElemParams = CreateMessage.createLocationMessage(
                description: params['desc'],
                longitude: params['longitude'],
                latitude: params['latitude'],
                priority: params['priority'],
                userID: sendToUserID);
            messageElem = timeweb!.createLocationMessage(createElemParams);
            break;
          }
        case "mergeMessage":
          {
            final createElemParams = CreateMessage.createMereMessage(
                userID: sendToUserID,
                messageList: params['webMessageInstanceList'],
                title: params['title'],
                priority: params['priority'],
                abstractList: params['abstractList'],
                compatibleText: params['compatibleText']);
            messageElem = timeweb!.createMergerMessage(createElemParams);
            break;
          }
        case "forwardMessage":
          {
            final createElemParams = CreateMessage.createForwardMessage(
              message: parse(params['webMessageInstance']),
              userID: sendToUserID,
              priority: params['priority'],
            );
            messageElem = timeweb!.createForwardMessage(createElemParams);
            break;
          }
        case "video":
          {
            final progressCallback = allowInterop((progress) async {
              final messageInstance =
                  await V2TIMMessage.convertMessageFromWebToDart(messageElem);
              if (messageListener != null) {
                messageListener!.onSendMessageProgress(
                    V2TimMessage.fromJson(messageInstance), progress as int);
              }
            });
            final createElemParams = CreateMessage.createVideoMessage(
                userID: sendToUserID,
                file: params['file'],
                convType: convType,
                progressCallback: progressCallback,
                priority: params['priority']);
            messageElem = timeweb!.createVideoMessage(createElemParams);
            break;
          }
        case "file":
          {
            final progressCallback = allowInterop((progress) async {
              final messageInstance =
                  await V2TIMMessage.convertMessageFromWebToDart(messageElem);
              if (messageListener != null) {
                messageListener!.onSendMessageProgress(
                    V2TimMessage.fromJson(messageInstance), progress as int);
              }
            });
            final createElemParams = CreateMessage.createFileMessage(
                userID: sendToUserID,
                file: params['file'],
                convType: convType,
                progressCallback: progressCallback,
                priority: params['priority']);
            messageElem = timeweb!.createFileMessage(createElemParams);
            break;
          }
      }
      final res = await promiseToFuture(timeweb!.sendMessage(messageElem,
          mapToJSObj({"onlineUserOnly": params['onlineUserOnly']})));
      final code = res.code;
      if (code == 0) {
        final message = res.data.message;
        final formatedMessage =
            await V2TIMMessage.convertMessageFromWebToDart(message);
        return CommonUtils.returnSuccess<F>(formatedMessage);
      } else {
        return CommonUtils.returnErrorForValueCb<F>('发送消息失败');
      }
    } catch (error) {
      return CommonUtils.returnErrorForValueCb<F>(error);
    }
  }

  Future<T> sendTextMessage<T, F>(Map<String, dynamic> params) async {
    return await sendMessage<T, F>(type: 'text', params: params);
  }

  Future<T> sendCustomMessage<T, F>(Map<String, dynamic> params) async {
    return await sendMessage<T, F>(type: 'custom', params: params);
  }

  static List<Object> generateDartListObject(List<dynamic> params) =>
      List.from(params);

  Future<T> sendImageMessage<T, F>(Map<String, dynamic> params) async {
    String? mimeType = mime(Path.basename(params['fileName']));
    final fileContent = generateDartListObject(params['fileContent']);
    params['file'] = html.File(
        fileContent, params['fileName'] as String, {'type': mimeType});
    return await sendMessage<T, F>(type: 'image', params: params);
  }

  Future<T> sendVideoMessage<T, F>(Map<String, dynamic> params) async {
    String? mimeType = mime(Path.basename(params['fileName']));
    final fileContent = generateDartListObject(params['fileContent']);
    params['file'] = html.File(
        fileContent, params['fileName'] as String, {'type': mimeType});
    return await sendMessage<T, F>(type: 'video', params: params);
  }

  Future<T> sendFileMessage<T, F>(Map<String, dynamic> params) async {
    String? mimeType = mime(Path.basename(params['fileName']));
    final fileContent = generateDartListObject(params['fileContent']);
    params['file'] = html.File(
        fileContent, params['fileName'] as String, {'type': mimeType});
    return await sendMessage<T, F>(type: 'file', params: params);
  }

  // web不支持
  Future<V2TimValueCallback<V2TimMessage>> sendSoundMessage() async {
    return CommonUtils.returnErrorForValueCb<V2TimMessage>(
        "getTotalUnreadMessageCount feature does not exist on the web");
  }

  Future<T> sendTextAtMessage<T, F>(Map<String, dynamic> params) async {
    return await sendMessage<T, F>(type: 'textAt', params: params);
  }

  Future<T> sendFaceMessage<T, F>(Map<String, dynamic> params) async {
    return await sendMessage<T, F>(type: 'face', params: params);
  }

  Future<T> sendLocationMessage<T, F>(Map<String, dynamic> params) async {
    return await sendMessage<T, F>(type: 'location', params: params);
  }

  Future<T> sendMergerMessage<T, F>(Map<String, dynamic> params) async {
    log(params);
    return await sendMessage<T, F>(type: 'mergeMessage', params: params);
  }

  Future<T> sendForwardMessage<T, F>(Map<String, dynamic> params) async {
    return await sendMessage<T, F>(type: 'forwardMessage', params: params);
  }

  Future<dynamic> reSendMessage(Map<String, dynamic> params) async {
    try {
      final res = await promiseToFuture(
          timeweb!.reSendMessage(parse(params['webMessageInstance'])));
      final code = res.code;
      if (code == 0) {
        final message = res.data.message;
        final formatedMessage =
            await V2TIMMessage.convertMessageFromWebToDart(message);
        return CommonUtils.returnSuccess<V2TimMessage>(formatedMessage);
      } else {
        return CommonUtils.returnErrorForValueCb<V2TimMessage>('重发失败!');
      }
    } catch (error) {
      return CommonUtils.returnErrorForValueCb<V2TimMessage>(error);
    }
  }

  Future<V2TimCallback> setLocalCustomData() async {
    return CommonUtils.returnError(
        "setLocalCustomData feature does not exist on the web");
  }

  Future<V2TimCallback> setLocalCustomInt() async {
    return CommonUtils.returnError(
        "setLocalCustomInt feature does not exist on the web");
  }

  Future<V2TimCallback> setCloudCustomData() async {
    return CommonUtils.returnError(
        "setCloudCustomData feature does not exist on the web");
  }

  Future<dynamic> getMessageList(Object getMsgListParams) async {
    try {
      final res =
          await promiseToFuture(timeweb!.getMessageList(getMsgListParams));
      final code = res.code;
      if (code == 0) {
        final List messageList = res.data.messageList;
        final historyMessageListPromises = messageList.map((element) async {
          final responses =
              await V2TIMMessage.convertMessageFromWebToDart(element);
          return responses;
        }).toList();
        final historyMessageList =
            await Future.wait(historyMessageListPromises);
        return CommonUtils.returnSuccess<List<V2TimMessage>>(
            historyMessageList);
      } else {
        return CommonUtils.returnError("获取历史消息失败");
      }
    } catch (error) {
      return CommonUtils.returnError(error);
    }
  }

  Future<dynamic> getC2CHistoryMessageList(params) async {
    final getMessageListParams = GetMessageList.formateParams(params);
    return await getMessageList(getMessageListParams);
  }

  Future<dynamic> getGroupHistoryMessageList(params) async {
    final getMessageListParams = GetMessageList.formateParams(params);
    return await getMessageList(getMessageListParams);
  }

  Future<Map<String, dynamic>> getHistoryMessageList(params) async {
    final getMessageListParams = GetMessageList.formateParams(params);
    if (getMessageListParams == null) {
      return {
        'desc': "receiver and groupID cannot set at the same time",
        'data': []
      };
    }
    return await getMessageList(getMessageListParams);
  }

  Future<dynamic> getHistoryMessageListWithoutFormat() async {
    return {
      "code": 0,
      "desc":
          "getHistoryMessageListWithoutFormat feature does not exist on the web"
    };
    // return CommonUtils.returnErrorForValueCb<LinkedHashMap<dynamic, dynamic>>(
    //     "getHistoryMessageListWithoutFormat feature does not exist on the web");
  }

  Future<dynamic> revokeMessage(params) async {
    try {
      final res = await promiseToFuture(
          timeweb!.revokeMessage(params['webMessageInstatnce']));
      final code = res.code;
      if (code == 0) {
        return CommonUtils.returnSuccessForCb(jsToMap(res.data));
      } else {
        return CommonUtils.returnError("撤回消息失败");
      }
    } catch (error) {
      return CommonUtils.returnError(error);
    }
  }

  Future<dynamic> markMsgReaded(markMsgReadedParams) async {
    try {
      final res =
          await promiseToFuture(timeweb!.setMessageRead(markMsgReadedParams));
      if (res.code == 0) {
        return CommonUtils.returnSuccessForCb(jsToMap(res.data));
      } else {
        return CommonUtils.returnError('设置已读失败');
      }
    } catch (error) {
      return CommonUtils.returnError(error);
    }
  }

  Future<V2TimCallback> markC2CMessageAsRead(params) async {
    final markMsgReadedParams =
        mapToJSObj({"conversationID": 'C2C' + params['userID']});
    return await markMsgReaded(markMsgReadedParams);
  }

  Future<dynamic> markGroupMessageAsRead(params) async {
    final markMsgReadedParams =
        mapToJSObj({"conversationID": 'GROUP' + params['groupID']});
    return await markMsgReaded(markMsgReadedParams);
  }

  Future<V2TimCallback> deleteMessageFromLocalStorage() async {
    return CommonUtils.returnError(
        "deleteMessageFromLocalStorage feature does not exist on the web");
  }

  Future<dynamic> insertGroupMessageToLocalStorage() async {
    return CommonUtils.returnErrorForValueCb<V2TimMessage>(
        "insertGroupMessageToLocalStorage feature does not exist on the web");
  }

  Future<dynamic> insertC2CMessageToLocalStorage() async {
    return CommonUtils.returnErrorForValueCb<V2TimMessage>(
        "insertGroupMessageToLocalStorage feature does not exist on the web");
  }

  Future<V2TimCallback> clearC2CHistoryMessage() async {
    return CommonUtils.returnError(
        "clearC2CHistoryMessage feature does not exist on the web");
  }

  Future<V2TimCallback> clearGroupHistoryMessage() async {
    return CommonUtils.returnError(
        "clearGroupHistoryMessage feature does not exist on the web");
  }

  Future<dynamic> getC2CReceiveMessageOpt() async {
    return CommonUtils.returnErrorForValueCb<List<V2TimReceiveMessageOptInfo>>(
        "getC2CReceiveMessageOpt feature does not exist on the web");
  }

  Future<dynamic> searchLocalMessages() async {
    return CommonUtils.returnErrorForValueCb<V2TimMessageSearchResult>(
        "searchLocalMessages feature does not exist on the web");
  }

  Future<dynamic> findMessages() async {
    return CommonUtils.returnErrorForValueCb<List<V2TimMessage>>(
        "findMessages feature does not exist on the web");
  }

  Future<V2TimCallback> setC2CReceiveMessageOpt() async {
    return CommonUtils.returnError(
        "setC2CReceiveMessageOpt feature does not exist on the web");
  }

  // webMessageInstanceList 这个参数web独有其中元素是web端的message实例
  Future<dynamic> deleteMessages(params) async {
    try {
      final res = await promiseToFuture(
          timeweb!.deleteMessages(params['webMessageInstanceList']));
      if (res.code == 0) {
        return CommonUtils.returnSuccess(jsToMap(res.data));
      } else {
        return CommonUtils.returnError('delete msg failed');
      }
    } catch (error) {
      log(error);
      return CommonUtils.returnError(error);
    }
  }

  Future<dynamic> setGroupReceiveMessageOpt(params) async {
    try {
      final remindTypeOpt = {
        'groupID': params['groupID'],
        'messageRemindType':
            GroupRecvMsgOpt.convertMsgRecvOptToWeb(params['opt'])
      };
      final res = await promiseToFuture(
          timeweb!.setMessageRemindType(mapToJSObj(remindTypeOpt)));
      if (res.code == 0) {
        return CommonUtils.returnSuccessForCb(jsToMap(res.data));
      } else {
        return CommonUtils.returnError('set group recv msg opt failed');
      }
    } catch (error) {
      return CommonUtils.returnError(error);
    }
  }

  static final _messageReadReceiptHandler = allowInterop((dynamic responses) {
    final List messageData = responses.data;
    final readedList = messageData
        .map((item) =>
            V2TimMessageReceipt(userID: item.from, timestamp: item.time))
        .toList();
    messageListener!.onRecvC2CReadReceipt(readedList);
  });

  static final _messageRevokedHandler = allowInterop((dynamic responses) {
    final List messageData = responses.data;
    for (var element in messageData) {
      final msgID = element.ID;
      messageListener!.onRecvMessageRevoked(msgID);
    }
  });

  static final _reciveNewMesageHandler = allowInterop((dynamic responses) {
    final List messageList = responses.data;
    for (var messageItem in messageList) {
      loop() async {
        final formatedMessage =
            await V2TIMMessage.convertMessageFromWebToDart(messageItem);
        messageListener!
            .onRecvNewMessage(V2TimMessage.fromJson(formatedMessage));
      }

      loop();
    }
  });

  void addAdvancedMsgListener(V2TimAdvancedMsgListener listener) {
    messageListener = listener;
    submitMessageReadReceipt();
    submitMessageRevoked();
    submitRecvNewMessage();
  }

  void removeAdvancedMsgListener() {
    timeweb!.off(EventType.MESSAGE_READ_BY_PEER, _messageReadReceiptHandler);
    timeweb!.off(EventType.MESSAGE_REVOKED, _messageRevokedHandler);
    timeweb!.off(EventType.MESSAGE_RECEIVED, _reciveNewMesageHandler);
  }

  static void submitMessageReadReceipt() {
    timeweb!.on(EventType.MESSAGE_READ_BY_PEER, _messageReadReceiptHandler);
  }

  static void submitMessageRevoked() {
    timeweb!.on(EventType.MESSAGE_REVOKED, _messageRevokedHandler);
  }

  static void submitRecvNewMessage() {
    timeweb!.on(EventType.MESSAGE_RECEIVED, _reciveNewMesageHandler);
  }
}
